import logging
import concurrent.futures
import random
import bisect
import translators as ts
from easytranslator.settings import SUPPORTED_LANGUAGES, STABLE_TS, LANG_MAP
from requests.exceptions import HTTPError
import itertools
from typing import Optional, List, Dict


class EasyTranslatorError(Exception):
    pass

class EasyTranslator:
    TIMEOUT = 10
    SLEEP_SECONDS = 0
    UPDATE_SESSION_AFTER_FREQ = 1000
    UPDATE_SESSION_AFTER_SECONDS = 1500

    def __init__(self, translators=[], top_n=3):
        stable_ts = translators if translators else STABLE_TS
        self.translators: List[dict] = sorted(stable_ts, key=lambda x: x["priority"])
        self.top_translators = self.translators[:top_n]
        self.remaining_translators = self.translators[top_n:]

    def get_translator(self, translator_id: str) -> Optional[Dict]:
        # for t in self.translators:
        #     if t["id"] == t_id:
        #         return t
        # return None
        return next((t for t in self.translators if t["id"] == translator_id), None)

    def get_supported_language(self) -> List:
        return SUPPORTED_LANGUAGES

    def set_priority(self, translator_id: str, priority: int) -> List[Dict]:
        translator = self.get_translator(translator_id)
        if not translator:
            raise EasyTranslatorError(f"Translator {translator_id} not supported")
        translator["priority"] = priority
        return self.translators

    def adjust_translator(self, translator, offset):
        logging.info("Adjusting translator %s by offset %s", translator['id'],offset)
        translator["priority"] += offset
        bisect.insort(self.translators, translator, key=lambda x: x["priority"])

    def translate_with_translator(
        self,
        translator: dict,
        text: str,
        dest_lang: str,
        src_lang: str = "auto",
        proxies: List = [],
    ) -> Dict:

        translator_id = translator["id"]
        translator_lang_map = LANG_MAP[translator_id]

        if dest_lang not in translator_lang_map:
            return {
                "translated_text": None,
                "status": "error",
                "error_info": f"{dest_lang} is not in {translator_id}'s Lang Map, Please check the SUPPORTED_LANGUAGES",
            }
        try:
            logging.info("Translation using %s: %s", translator_id,text)
            result = ts.translate_text(
                text,
                translator=translator_id,
                from_language=src_lang,
                to_language=translator_lang_map[dest_lang],
                timeout=self.TIMEOUT,
                proxies=proxies,
                sleep_seconds=self.SLEEP_SECONDS,
                update_session_after_freq=self.UPDATE_SESSION_AFTER_FREQ,
                update_session_after_seconds=self.UPDATE_SESSION_AFTER_SECONDS,
            )

            if result:
                self.adjust_translator(translator, 1)
                return {"translated_text": result, "status": "success"}
        except ConnectionError as e:
            self.adjust_translator(translator, -2)
            logging.warning("Failed to translate using %s: ConnectionError, %s", translator_id,e)
        # except HTTPError as e:
        #     self.adjust_translator(translator, -1)
        #     assert(f"Failed to translate using {translator_id} {e}: HTTPError")

        except TimeoutError as e:
            self.adjust_translator(translator, -3)
            logging.warning("Failed to translate using %s: TimeoutError, %s", translator_id,e)
        except Exception as e:
            self.adjust_translator(translator, -3)
            logging.warning("Failed to translate using %s: Exception, %s", translator_id,e)

        return {
            "translated_text": None,
            "status": "error",
            "error_info": f"Failed to translate using {translator_id}",
        }

    def translate(
        self, text: str, dest_lang: str, src_lang: str = "auto", proxies: List = []
    ) -> Dict:
        if not text or not dest_lang:
            return {
                "translated_text": None,
                "status": "error",
                "error_info": "Text or dest_lang is empty",
            }

        translator = random.choice(self.top_translators)
        if dest_lang not in LANG_MAP[translator["id"]]:
            return {
                "translated_text": None,
                "status": "error",
                "error_info": f"{dest_lang} is not in {translator['id']}'s Lang Map, Please check the SUPPORTED_LANGUAGES",
            }

        result = self.translate_with_translator(
            translator, text, dest_lang, src_lang, proxies
        )
        if result.get("status") == "success":
            return result

        for translator in self.remaining_translators:
            result = self.translate_with_translator(
                translator, text, dest_lang, src_lang, proxies
            )
            if result.get("status") == "success":
                return result

        logging.error("All services failed to translate!")
        return {
            "translated_text": None,
            "status": "error",
            "error_info": "All services failed to translate!",
        }

    def translate_batch(
        self, text_list: list, dest_lang: str, src_lang: str = "auto"
    ) -> Dict:
        with concurrent.futures.ThreadPoolExecutor() as executor:
            results = {
                text: translation
                for text, translation in zip(
                    text_list,
                    executor.map(
                        self.translate,
                        text_list,
                        itertools.repeat(dest_lang),
                        itertools.repeat(src_lang),
                    ),
                )
            }
        return results



# test_easy_translate.py
import pytest
from unittest.mock import MagicMock, patch
from easytranslator import EasyTranslator
from easytranslator.main import EasyTranslatorError
from easytranslator.settings import SUPPORTED_LANGUAGES, STABLE_TS

@pytest.fixture
def et():
    return EasyTranslator()
    
def test_init(et):
    assert len(et.translators) == len(STABLE_TS)
    assert len(et.top_translators) == 3
    assert len(et.remaining_translators) == len(STABLE_TS)-3
    

def test_translate(et):
    result = et.translate('Hello', dest_lang='Chinese Simplified')
    assert result['status'] == 'success'
    assert result['translated_text'] in ['您好','你好']

    result = et.translate('World', dest_lang='Chinese Simplified')
    assert result['status'] == 'success'
    assert result['translated_text'] in ['世界']

    result = et.translate('', dest_lang='Chinese Simplified')
    assert result['status'] == 'error'

    result = et.translate('World', dest_lang='zh-CN')
    assert result['status'] == 'error'

def test_translate_batch(et):
    results = et.translate_batch(['Hello', 'World'], dest_lang='Chinese Simplified', src_lang='en')
    assert len(results) == 2
    print(results)
    assert results['Hello']['status'] == 'success'
    assert results['World']['status'] == 'success' 
    assert results['Hello']['translated_text'] in ['您好','你好']
    assert results['World']['translated_text'] in ['世界']

def test_adjust_translator(et):
    translator = et.translators[0]
    old_priority = translator["priority"]
    
    et.adjust_translator(translator, 1)
    assert translator["priority"] == old_priority + 1
    
    et.adjust_translator(translator, -old_priority)
    assert translator["priority"] == 1

def test_get_translator(et):
    assert et.get_translator('google')
    assert not et.get_translator('haha')

def test_get_supported_language(et):
    assert len(et.get_supported_language()) == len(SUPPORTED_LANGUAGES)

def test_set_priority(et):
    et.set_priority('myMemory', 100)
    assert et.get_translator('myMemory').get('priority') == 100

    with pytest.raises(EasyTranslatorError) as excinfo:
        et.set_priority('haha', 100)
    assert str(excinfo.value) == "Translator haha not supported"




# STABLE_TS = {
#     "myMemory": {"name": "MyMemory", "priority": 20},
#     "alibaba": {"name": "Alibaba", "priority": 10},
#     "baidu": {"name": "Baidu", "priority": 10},
#     "modernMt": {"name": "ModernMT", "priority": 20},
#     "iciba": {"name": "Iciba", "priority": 15},
#     "google": {"name": "Google", "priority": 4},
#     "bing": {"name": "Bing", "priority": 5},
#     "lingvanex": {"name": "Lingvanex", "priority": 20},
#     # "yandex": {"name": "Yandex", "priority": 20},
#     "itranslate": {"name": "Itranslate", "priority": 20},
#     "sysTran": {"name": "SysTran", "priority": 20},
#     "argos": {"name": "ArgoS", "priority": 20},
#     "reverso": {"name": "Reverso", "priority": 20},
#     "deepl": {"name": "DeepL", "priority": 5},
#     "cloudTranslation": {"name": "Cloud Translation", "priority": 10},
#     "qqTranSmart": {"name": "QQ Translate Smart", "priority": 5},
#     "translateCom": {"name": "Translate Com", "priority": 15},
#     "sogou": {"name": "Sogou", "priority": 13},
#     "qqFanyi": {"name": "QQ Fanyi", "priority": 10},
#     "papago": {"name": "Papago", "priority": 15},
#     "youdao": {"name": "Youdao", "priority": 15},
#     "iflyrec": {"name": "iFlyrec", "priority": 30},
#     "caiyun": {"name": "Caiyun", "priority": 15},
# }
STABLE_TS = [
    {"id":"myMemory", "name": "MyMemory", "priority": 20},
    {"id":"alibaba", "name": "Alibaba", "priority": 10},
    {"id":"baidu", "name": "Baidu", "priority": 10},
    {"id":"modernMt", "name": "ModernMT", "priority": 20},
    {"id":"iciba", "name": "Iciba", "priority": 15},
    {"id":"google", "name": "Google", "priority": 4},
    {"id":"bing", "name": "Bing", "priority": 5},
    {"id":"lingvanex", "name": "Lingvanex", "priority": 20},
    # {"id":"yandex", "name": "Yandex", "priority": 20},
    {"id":"itranslate", "name": "Itranslate", "priority": 20},
    {"id":"sysTran", "name": "SysTran", "priority": 20},
    {"id":"argos", "name": "ArgoS", "priority": 20},
    {"id":"reverso", "name": "Reverso", "priority": 20},
    {"id":"deepl", "name": "DeepL", "priority": 5},
    {"id":"cloudTranslation", "name": "Cloud Translation", "priority": 10},
    {"id":"qqTranSmart", "name": "QQ Translate Smart", "priority": 5},
    {"id":"translateCom", "name": "Translate Com", "priority": 15},
    {"id":"sogou", "name": "Sogou", "priority": 13},
    {"id":"qqFanyi", "name": "QQ Fanyi", "priority": 10},
    {"id":"papago", "name": "Papago", "priority": 15},
    {"id":"youdao", "name": "Youdao", "priority": 15},
    {"id":"iflyrec", "name": "iFlyrec", "priority": 30},
    {"id":"caiyun", "name": "Caiyun", "priority": 15},
]
SUPPORTED_LANGUAGES = [
    "English",
    "Chinese Simplified",
    "Chinese Traditional",
    "Russian",
    "Japanese",
    "Korean",
    "Czech",
    "Danish",
    "German",
    "Spanish",
    "French",
    "Indonesian",
    "Italian",
    "Hungarian",
    "Dutch",
    "Polish",
    "Portuguese",
    "Swedish",
    "Turkish",
]

LANG_MAP = {
    "myMemory": {
        "Czech": "cs-CZ",
        "Danish": "da-DK",
        "German": "de-DE",
        "English": "en-ZA",
        "Spanish": "es-US",
        "French": "fr-FR",
        "Hungarian": "hu-HU",
        "Indonesian": "id-ID",
        "Italian": "it-IT",
        "Japanese": "ja-JP",
        "Korean": "ko-KR",
        "Dutch": "nl-NL",
        "Polish": "pl-PL",
        "Portuguese": "pt-PT",
        "Russian": "ru-RU",
        "Swedish": "sv-SE",
        "Turkish": "tr-TR",
        "Chinese Simplified": "zh-CN",
        "Chinese Traditional": "zh-TW",
    },
    "alibaba": {
        "Czech": "cs",
        "Danish": "da",
        "German": "de",
        "English": "en",
        "Spanish": "es",
        "French": "fr",
        "Hungarian": "hu",
        "Indonesian": "id",
        "Italian": "it",
        "Japanese": "ja",
        "Korean": "ko",
        "Dutch": "nl",
        "Polish": "pl",
        "Portuguese": "pt",
        "Russian": "ru",
        "Swedish": "sv",
        "Turkish": "tr",
        "Chinese Simplified": "zh",
        "Chinese Traditional": "zh-tw",
    },
    "baidu": {
        "Czech": "cs",
        "German": "de",
        "English": "en",
        "Hungarian": "hu",
        "Indonesian": "id",
        "Italian": "it",
        "Dutch": "nl",
        "Polish": "pl",
        "Portuguese": "pt",
        "Russian": "ru",
        "Turkish": "tr",
        "Chinese Simplified": "zh",
    },
    "modernMt": {
        "Czech": "cs",
        "Danish": "da",
        "German": "de",
        "English": "en",
        "Spanish": "es-ES",
        "French": "fr",
        "Hungarian": "hu",
        "Indonesian": "id",
        "Italian": "it",
        "Japanese": "ja",
        "Korean": "ko",
        "Dutch": "nl",
        "Polish": "pl",
        "Portuguese": "pt-PT",
        "Russian": "ru",
        "Swedish": "sv",
        "Turkish": "tr",
        "Chinese Simplified": "zh-CN",
        "Chinese Traditional": "zh-TW",
    },
    "iciba": {
        "Czech": "cs",
        "Danish": "da",
        "German": "de",
        "English": "en",
        "Spanish": "es",
        "French": "fr",
        "Hungarian": "hu",
        "Indonesian": "id",
        "Italian": "it",
        "Japanese": "ja",
        "Korean": "ko",
        "Dutch": "nl",
        "Polish": "pl",
        "Portuguese": "pt",
        "Russian": "ru",
        "Swedish": "sv",
        "Turkish": "tr",
        "Chinese Simplified": "zh",
    },
    "google": {
        "Czech": "cs",
        "Danish": "da",
        "German": "de",
        "English": "en-US",
        "Spanish": "es",
        "French": "fr",
        "Hungarian": "hu",
        "Indonesian": "id",
        "Italian": "it",
        "Japanese": "ja",
        "Korean": "ko",
        "Dutch": "nl",
        "Polish": "pl",
        "Portuguese": "pt",
        "Russian": "ru",
        "Swedish": "sv",
        "Turkish": "tr",
        "Chinese Simplified": "zh-CN",
        "Chinese Traditional": "zh-TW",
    },
    "bing": {
        "Czech": "cs",
        "Danish": "da",
        "German": "de",
        "English": "en",
        "Spanish": "es",
        "French": "fr-CA",
        "Hungarian": "hu",
        "Indonesian": "id",
        "Italian": "it",
        "Japanese": "ja",
        "Korean": "ko",
        "Dutch": "nl",
        "Polish": "pl",
        "Portuguese": "pt-PT",
        "Russian": "ru",
        "Swedish": "sv",
        "Turkish": "tr",
        "Chinese Simplified": "zh-Hans",
        "Chinese Traditional": "zh-Hant",
    },
    "lingvanex": {
        "Czech": "cs_CZ",
        "Danish": "da_DK",
        "German": "de_DE",
        "English": "en_GB",
        "Spanish": "es_ES",
        "French": "fr_FR",
        "Hungarian": "hu_HU",
        "Indonesian": "id_ID",
        "Italian": "it_IT",
        "Japanese": "ja_JP",
        "Korean": "ko_KR",
        "Dutch": "nl_NL",
        "Polish": "pl_PL",
        "Portuguese": "pt_PT",
        "Russian": "ru_RU",
        "Swedish": "sv_SE",
        "Turkish": "tr_TR",
        "Chinese Simplified": "zh-Hans_CN",
    },
    "itranslate": {
        "Czech": "cs",
        "Danish": "da",
        "German": "de",
        "English": "en-US",
        "Spanish": "es-US",
        "French": "fr-FR",
        "Hungarian": "hu",
        "Indonesian": "id",
        "Italian": "it",
        "Japanese": "ja",
        "Korean": "ko",
        "Dutch": "nl",
        "Polish": "pl",
        "Portuguese": "pt-PT",
        "Russian": "ru",
        "Swedish": "sv",
        "Turkish": "tr",
        "Chinese Simplified": "zh-CN",
        "Chinese Traditional": "zh-TW",
    },
    "sysTran": {
        "Czech": "cs",
        "Danish": "da",
        "German": "de",
        "English": "en",
        "Spanish": "es",
        "French": "fr",
        "Hungarian": "hu",
        "Indonesian": "id",
        "Italian": "it",
        "Japanese": "ja",
        "Korean": "ko",
        "Dutch": "nl",
        "Polish": "pl",
        "Portuguese": "pt",
        "Russian": "ru",
        "Swedish": "sv",
        "Turkish": "tr",
        "Chinese Simplified": "zh",
    },
    "argos": {
        "Czech": "cs",
        "Danish": "da",
        "German": "de",
        "English": "en",
        "Spanish": "es",
        "French": "fr",
        "Hungarian": "hu",
        "Indonesian": "id",
        "Italian": "it",
        "Japanese": "ja",
        "Korean": "ko",
        "Dutch": "nl",
        "Polish": "pl",
        "Portuguese": "pt",
        "Russian": "ru",
        "Swedish": "sv",
        "Turkish": "tr",
        "Chinese Simplified": "zh",
    },
    "reverso": {
        "Czech": "cs",
        "Danish": "da",
        "German": "de",
        "English": "en",
        "Spanish": "es",
        "French": "fr",
        "Hungarian": "hu",
        "Indonesian": "id",
        "Italian": "it",
        "Japanese": "ja",
        "Korean": "ko",
        "Dutch": "nl",
        "Polish": "pl",
        "Portuguese": "pt",
        "Russian": "ru",
        "Swedish": "sv",
        "Turkish": "tr",
        "Chinese Simplified": "zh",
    },
    "deepl": {
        "Czech": "cs",
        "Danish": "da",
        "German": "de",
        "English": "en",
        "Spanish": "es",
        "French": "fr",
        "Hungarian": "hu",
        "Indonesian": "id",
        "Italian": "it",
        "Japanese": "ja",
        "Korean": "ko",
        "Dutch": "nl",
        "Polish": "pl",
        "Portuguese": "pt",
        "Russian": "ru",
        "Swedish": "sv",
        "Turkish": "tr",
        "Chinese Simplified": "zh",
    },
    "cloudTranslation": {
        "German": "de",
        "English": "en-us",
        "Spanish": "es",
        "French": "fr",
        "Indonesian": "id",
        "Italian": "it",
        "Japanese": "ja",
        "Korean": "ko",
        "Dutch": "nl",
        "Polish": "pl",
        "Portuguese": "pt",
        "Russian": "ru",
        "Turkish": "tr",
        "Chinese Simplified": "zh-cn",
        "Chinese Traditional": "zh-tw",
    },
    "qqTranSmart": {
        "Czech": "cs",
        "Danish": "da",
        "German": "de",
        "English": "en",
        "Spanish": "es",
        "French": "fr",
        "Hungarian": "hu",
        "Indonesian": "id",
        "Italian": "it",
        "Japanese": "ja",
        "Korean": "ko",
        "Dutch": "nl",
        "Polish": "pl",
        "Portuguese": "pt-PT",
        "Russian": "ru",
        "Swedish": "sv",
        "Turkish": "tr",
        "Chinese Simplified": "zh",
    },
    "translateCom": {
        "English": "en-gb",
        "French": "fr-ca",
        "German": "de",
        "Italian": "it",
        "Portuguese": "pt-br",
        "Russian": "ru",
        "Spanish": "es-la",
        "Japanese": "ja",
        "Chinese Simplified": "zh",
        "Chinese Traditional": "zh-TW",
        "Korean": "ko",
        "Polish": "pl",
        "Dutch": "nl",
        "Indonesian": "id",
        "Swedish": "sv",
        "Danish": "da",
        "Czech": "cs",
        "Turkish": "tr",
        "Hungarian": "hu",
    },
    "sogou": {
        "Polish": "pl",
        "Danish": "da",
        "German": "de",
        "Russian": "ru",
        "French": "fr",
        "Korean": "ko",
        "Dutch": "nl",
        "Czech": "cs",
        "Portuguese": "pt",
        "Japanese": "ja",
        "Swedish": "sv",
        "Spanish": "es",
        "Hungarian": "hu",
        "English": "en",
        "Italian": "it",
        "Chinese Simplified": "zh-CHS",
    },
    "qqFanyi": {
        "English": "en",
        "Chinese Simplified": "zh",
        "French": "fr",
        "Spanish": "es",
        "Italian": "it",
        "German": "de",
        "Turkish": "tr",
        "Russian": "ru",
        "Portuguese": "pt",
        "Indonesian": "id",
    },
    "papago": {
        "German": "de",
        "English": "en",
        "Spanish": "es",
        "French": "fr",
        "Indonesian": "id",
        "Italian": "it",
        "Japanese": "ja",
        "Korean": "ko",
        "Portuguese": "pt",
        "Russian": "ru",
        "Chinese Simplified": "zh-CN",
        "Chinese Traditional": "zh-TW",
    },
    "youdao": {
        "German": "de",
        "English": "en",
        "Spanish": "es",
        "French": "fr",
        "Indonesian": "id",
        "Japanese": "ja",
        "Korean": "ko",
        "Portuguese": "pt",
        "Russian": "ru",
        "Chinese Simplified": "zh-CHS",
    },
    "iflyrec": {
        "German": "de",
        "English": "en",
        "Spanish": "es",
        "French": "fr",
        "Italian": "it",
        "Japanese": "ja",
        "Korean": "ko",
        "Russian": "ru",
        "Chinese Simplified": "zh",
    },
    "caiyun": {
        "English": "en",
        "Spanish": "es",
        "French": "fr",
        "Japanese": "ja",
        "Korean": "ko",
        "Russian": "ru",
        "Chinese Simplified": "zh",
    },
}
